
#include "convert.h"

struct conversions {
	enum dimension dimension;
	enum unit unit;
	double multforSIunit;
	double addforSIunit;
} conversions[]={
	/* lengths */
	{ DIM_LENGTH,	UNIT_METER,		1.0,	0.0},
	{ DIM_LENGTH,	UNIT_KILOMETER,       1.0e+3,	0.0},
	{ DIM_LENGTH,	UNIT_DECIMETER,       1.0e-1,	0.0},
	{ DIM_LENGTH,	UNIT_CENTIMETER,      1.0e-2,	0.0},
	{ DIM_LENGTH,	UNIT_MILLIMETER,      1.0e-3,	0.0},
	{ DIM_LENGTH,	UNIT_MICROMETER,      1.0e-6,	0.0},
	{ DIM_LENGTH,	UNIT_NANOMETER,       1.0e-9,	0.0},
	{ DIM_LENGTH,	UNIT_PICOMETER,       1.0e-12,0.0},
	{ DIM_LENGTH,	UNIT_FEMTOMETER,      1.0e-15,0.0},
	{ DIM_LENGTH,	UNIT_FEET,		0.3048 ,0.0},
	{ DIM_LENGTH,	UNIT_INCH,		0.0254 ,0.0},
	{ DIM_LENGTH,	UNIT_YARD,		0.9144 ,0.0},
	{ DIM_LENGTH,	UNIT_FATHOM,		1.8288 ,0.0},
	{ DIM_LENGTH,	UNIT_MILE,		1609.344,0.0},
	{ DIM_LENGTH,	UNIT_NAUTICAL_MILE,	1852.0 ,0.0},
	/* volumes */
	{ DIM_VOLUME,	UNIT_METER_3,		1.0,	0.0},
	{ DIM_VOLUME,	UNIT_LITER,		1.0e-3, 0.0},
	{ DIM_VOLUME,	UNIT_CENTILITER,	1.0e-5, 0.0},
	{ DIM_VOLUME,	UNIT_MILLILITER,	1.0e-6, 0.0},
	{ DIM_VOLUME,	UNIT_MICROLITER,	1.0e-9, 0.0},
	{ DIM_VOLUME,	UNIT_HECTOLITER,	1.0e-1, 0.0},
	{ DIM_VOLUME,	UNIT_GALLON_US_LIQUID,3.785e-3, 0.0},
	{ DIM_VOLUME,	UNIT_GALLON_US_DRY,   4.405e-3, 0.0},
	{ DIM_VOLUME,	UNIT_GALLON_UK,       4.546e-3, 0.0},
	{ DIM_VOLUME,	UNIT_INCH_3,      16.387064e-6, 0.0},
	{ DIM_VOLUME,	UNIT_FOOT_3,         28.317e-3, 0.0},
	/* temperatures */
	{ DIM_TEMPERATURE,	UNIT_KELVIN,		1.0,	 0.0},
	{ DIM_TEMPERATURE,	UNIT_CELSIUS,		1.0,   273.15},
	{ DIM_TEMPERATURE,	UNIT_FAHRENHEIT, 0.555555,   255.372},
	/* Masses */
	{ DIM_MASS,	UNIT_KILOGRAM,	1.0,	0.0},
	{ DIM_MASS,	UNIT_GRAM,		1.0e-3,	0.0},
	{ DIM_MASS,	UNIT_MILLIGRAM,	1.0e-6,	0.0},
	{ DIM_MASS,	UNIT_MICROGRAM,	1.0e-9,	0.0},
	{ DIM_MASS,	UNIT_NANOGRAM,	1.0e-12,0.0},
	{ DIM_MASS,	UNIT_PICOGRAM,	1.0e-15,0.0},
	{ DIM_MASS,	UNIT_FEMTOGRAM,	1.0e-18,0.0},
	{ DIM_MASS,	UNIT_TON,		1.0e+3 ,0.0},
	{ DIM_MASS,	UNIT_POUND,		0.4536, 0.0},
	/* PRESSURE */
	{ DIM_PRESSURE,	UNIT_PASCAL,		1.0,    0.0},
	{ DIM_PRESSURE,	UNIT_BAR,		1.0e+5, 0.0},
	{ DIM_PRESSURE,	UNIT_ATM,	    1.01325e+5, 0.0},
	{ DIM_PRESSURE,	UNIT_PSI,	    6894.757,	0.0},
	{ DIM_PRESSURE,	UNIT_FOOT_SEAWATER, 3068.167, 0.0},
	{ DIM_PRESSURE,	UNIT_TORR,	     133.322,   0.0},
	/* time */
	{ DIM_TIME,	UNIT_SECOND,		1.0,	0.0},
	{ DIM_TIME,	UNIT_MILLISECOND,     1.0e-3, 0.0},
	{ DIM_TIME,	UNIT_MICROSECOND,	1.0e-6, 0.0},
	{ DIM_TIME,	UNIT_NANOSECOND,	1.0e-9, 0.0},
	{ DIM_TIME,	UNIT_PICOSECOND,	1.0e-12,0.0},
	{ DIM_TIME,	UNIT_FEMTOSECOND,	1.0e-15,0.0},
	{ DIM_TIME,	UNIT_ATTOSECOND,	1.0e-18,0.0},
	{ DIM_TIME,	UNIT_MINUTE,	       60.0,	0.0},
	{ DIM_TIME,	UNIT_HOUR,	     3600.0,    0.0},
	{ DIM_TIME,	UNIT_DAY,	    86400.0,    0.0},
	{ DIM_TIME,	UNIT_WEEK,	   604800.0,	0.0},
	/* eof marker */
	{ DIM_EOF,	UNIT_EOF,		0.0,	0.0}
};


double convert(enum dimension dim,enum unit from,enum unit to,double value,int *error) {

	int x,err;
	
	err=1;
	for(x=0;conversions[x].dimension!=DIM_EOF;x++) {
		if (conversions[x].dimension==dim &&
		    conversions[x].unit     ==from) {
		    	value*=conversions[x].multforSIunit;
		    	value+=conversions[x].addforSIunit;
			goto fromok;
		}
	}
	goto errout;
	
	fromok:
	/* Value should now be SI */
	err=2;
	for(x=0;conversions[x].dimension!=DIM_EOF;x++) {
		if (conversions[x].dimension==dim &&
		    conversions[x].unit     ==to) {
		    	value-=conversions[x].addforSIunit;
		    	value/=conversions[x].multforSIunit;
			goto took;
		}
	}
	goto errout;

	took:
	err=0;

	errout:
	if (error) *error=err;
	if (err) value=-1.0;

	return value;
}
