
#undef DEFAULT_TO_IMPERIAL

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <unistd.h>

#define CB_CONTENT 0
#define CB_OPEN    1
#define CB_CLOSE   2
#define CB_OPTION  3

#include "convert.h"

struct divepoint {
	double time,depth;
	char *alarm;
};

struct dive {
	char date[11],time[6];
	double delta;
	double divetime;
	struct divepoint *divepoint;
	int divepoints,divepointalloc;
} divestruct={
	 "2000-01-01","00:00",
	 0.0,
	 0.0,
	 NULL,
	 0,0
};

struct react {
	int mode;
	unsigned char *path,*tex;
	void *priv;
	int (*runme)(struct react *self,unsigned char *content);
};

int texme(struct react *r,unsigned char *content) {
	printf("\\%s\t{%s}\n",r->tex,content);
	return 0;
}

enum convert_index { CV_LENGTH, CV_TEMP, CV_VOL, CV_PRESS, CV_OFP, CV_END};

struct converts {
	enum dimension dim;
	enum unit from,to;
};


struct converts myconverts[CV_END]={
	{ DIM_LENGTH,		UNIT_METER,	UNIT_METER},
	{ DIM_TEMPERATURE,	UNIT_KELVIN,	UNIT_CELSIUS},
	{ DIM_VOLUME,		UNIT_METER_3,	UNIT_LITER},
	{ DIM_PRESSURE,		UNIT_PASCAL,	UNIT_BAR},
	{ DIM_TIME,		UNIT_SECOND,	UNIT_MINUTE}
};

struct texts {
	char *lengthunit;
	char *language;
};

struct texts texts_metric={
	"m",	/* meter */
	"german"	/* language preset */
};

struct texts texts_imperial={
	"ft",	/* foot */
	"english"
};

struct texts *mytexts=&texts_metric;

void go_imperial(void){
	mytexts=&texts_imperial;
	myconverts[CV_LENGTH].to=UNIT_FOOT;
	myconverts[CV_TEMP]  .to=UNIT_FAHRENHEIT;
	myconverts[CV_VOL]   .to=UNIT_FOOT_3;
	myconverts[CV_PRESS] .to=UNIT_PSI;
	myconverts[CV_OFP]   .to=UNIT_MINUTE;
}

void go_metric(void){
	mytexts=&texts_metric;
	myconverts[CV_LENGTH].to=UNIT_METER;
	myconverts[CV_TEMP]  .to=UNIT_CELSIUS;
	myconverts[CV_VOL]   .to=UNIT_LITER;
	myconverts[CV_PRESS] .to=UNIT_BAR;
	myconverts[CV_OFP]   .to=UNIT_MINUTE;
}


int texmeconv(struct react *r,unsigned char *content) {

	double value;

	struct converts *conv=r->priv;

	sscanf(content,"%lf",&value);
	value=convert(conv->dim,conv->from,conv->to,value,NULL);

	printf("\\%s\t{%.2f}\n",r->tex,value);
	return 0;
}

int texpriv(struct react *r,unsigned char *content) {
	printf("\\%s\t{%s}\n",r->tex,(unsigned char *)r->priv);
	return 0;
}

int checkunits(struct react *r,unsigned char *content) {
	if (strcasecmp(content,"si")!=0) {
		fprintf(stderr,"AIEE - non SI input file.\n");
		exit(1);
	}
	return 0;
}

int sprintpad(struct react *r,unsigned char *content) {
	unsigned char *target=r->priv;
	int len=strlen(content);
	int wantlen=atoi(r->tex+1);
	if (len>wantlen) len=wantlen;
	while(len<wantlen) {*target++=*r->tex;wantlen--;}
	strncpy(target,content,len);
	return 0;
}

int getdouble(struct react *r,unsigned char *content) {
	double *target=r->priv;
	sscanf(content,"%lf",target);
	return 0;
}

int initdepth(struct react *r,unsigned char *content) {
	if (divestruct.divepoints&&divestruct.divepoint) {
		int x;
		for(x=0;x<divestruct.divepointalloc;x++) {
			if (divestruct.divepoint[x].alarm) {
				free(divestruct.divepoint[x].alarm);
				divestruct.divepoint[x].alarm=NULL;
			}
		}
	}
	if (divestruct.divepoint) free(divestruct.divepoint);
	divestruct.divepoint=NULL;
	divestruct.divepoints=0;
	divestruct.divepointalloc=0;
	divestruct.divetime=0;
	return 0;
}

int extendarray(void) {
	int old,x;

	if (divestruct.divepoints>=divestruct.divepointalloc) {
		old=divestruct.divepointalloc;
		divestruct.divepointalloc+=1024;
		divestruct.divepoint=realloc(divestruct.divepoint,divestruct.divepointalloc*sizeof(divestruct.divepoint[0]));
		if (!divestruct.divepoint) {
			fprintf(stderr,"AIEE - too many points when going %d -> %d.\n",divestruct.divepoints,divestruct.divepointalloc);
			return 1;
		}
		for(x=old;x<divestruct.divepointalloc;x++) {
			divestruct.divepoint[x].alarm=NULL;
		}
	}
	return 0;
}

int getdepth(struct react *r,unsigned char *content) {

	double value;
	struct converts *conv=r->priv;

	if (extendarray()) return 1;

	divestruct.divepoint[divestruct.divepoints].time=divestruct.divetime;
	sscanf(content,"%lf",&value);
	value=convert(conv->dim,conv->from,conv->to,value,NULL);
	divestruct.divepoint[divestruct.divepoints].depth=value;

	divestruct.divepoints++;
	divestruct.divetime+=divestruct.delta;
	return 0;
}

int getalarm(struct react *r,unsigned char *content) {

	if (extendarray()) return 1;

	divestruct.divepoint[divestruct.divepoints].alarm=strdup(content);
	return 0;
}

int showdepth(struct react *r,unsigned char *content) {
	int x,xx;
	double maxdepth,maxtime,maxdepthscale;
	
#define NUMSTEPS 9
	int steps[NUMSTEPS]={1,2,5,10,20,50,100,200,500};
	
	maxdepth=maxtime=-1.0;
	for(x=0;x<divestruct.divepoints;x++) {
		if (divestruct.divepoint[x].time>maxtime)
			maxtime=divestruct.divepoint[x].time;
		if (divestruct.divepoint[x].depth>maxdepth)
			maxdepth=divestruct.divepoint[x].depth;
	}
	maxdepthscale=(int)(maxdepth*1.01+0.5);
	printf( "\\ProfileGraph{\n");
	for(x=0;x<NUMSTEPS-1;x++) {
		if (maxtime/60.0/steps[x]<20.0) break;
	}
	for(xx=0;xx*60<maxtime;xx+=steps[x]) {
		printf("  \\put(%.3f,2.5){\\line(0,1){45}}\n",
			xx*60.0*90.0/maxtime+5.0);
		printf("  \\put(%.3f,48.0){\\footnotesize %d}\n",
			xx*60.0*90.0/maxtime+5.0-1.0,xx);
	}
	printf("  \\put(95,2.5){\\line(0,1){45}}\n");
	for(x=0;x<NUMSTEPS-1;x++) {
		if (maxdepthscale/steps[x]<10.0) break;
	}
	maxdepthscale=((int)(maxdepthscale/steps[x])+1)*steps[x]+1e-9;
	for(xx=0;xx<maxdepthscale;xx+=steps[x]) {
		printf("  \\put(5,%.3f){\\line(1,0){90}}\n",
			45.0-xx*45.0/maxdepthscale+2.5);
		printf("  \\put(1,%.3f){\\footnotesize %d}\n",
			45.0-xx*45.0/maxdepthscale-1.0+2.5,xx);
		printf("  \\put(96,%.3f){\\footnotesize %d}\n",
			45.0-xx*45.0/maxdepthscale-1.0+2.5,xx);
	}
	printf("  \\put(5,2.5){\\line(1,0){90}}\n");
	for(x=0;x<divestruct.divepoints;x++) {
		printf("  \\put(%.3f,%.3f){\\circle{1}{\\tiny %s}}\n",
			divestruct.divepoint[x].time *90.0/maxtime+5.0,
			45.0-divestruct.divepoint[x].depth*45.0 /maxdepthscale+2.5,
			(divestruct.divepoint[x].alarm ? divestruct.divepoint[x].alarm : ""));
	}
	printf( "}\n"
	        "\n"
	        "\\Duration     {%.1f min}\n"
	        "\\Maxdepth {%.1f\\,%s}\n"
	        ,maxtime/60.0,maxdepth+0.05,mytexts->lengthunit);

	{
		int endhr=atoi(divestruct.time);
		int endmin=atoi(divestruct.time+3);
		
		endmin+=maxtime/60.0;
		while(endmin>=60) {endmin-=60;endhr++;}
		while(endhr>24) endhr-=24;
		
		printf("\\Timeend  {%02d:%02d}\n",endhr,endmin);
	}

	return 0;
}

int enddive(struct react *r,unsigned char *content) {
	printf(	"\\begin{divelog}\n"
		"Notes go here.\n"
		"\\end{divelog}\n\n\n");
	return 0;
}

struct react reactions[]={
	 {CB_CONTENT,">PROFILE>UNITS",NULL,NULL,checkunits },
	 {CB_CONTENT,">PROFILE>DEVICE>MODEL","CDModel",NULL,texme },
	 {CB_CONTENT,">PROFILE>ADDINFO>PERSINFO","CDPersonal",NULL,texme },
	 {CB_CONTENT,">PROFILE>ADDINFO>SERIALID","CDSerial",NULL,texme },
	 {CB_OPEN   ,">PROFILE>REPGROUP>DIVE",NULL,NULL,initdepth },
	 {CB_CONTENT,">PROFILE>REPGROUP>DIVE>DATE>DAY","02",&divestruct.date[8],sprintpad },
	 {CB_CONTENT,">PROFILE>REPGROUP>DIVE>DATE>MONTH","02",&divestruct.date[5],sprintpad },
	 {CB_CONTENT,">PROFILE>REPGROUP>DIVE>DATE>YEAR"," 4",&divestruct.date[0],sprintpad },
	 {CB_CLOSE  ,">PROFILE>REPGROUP>DIVE>DATE","Date",&divestruct.date[0],texpriv },
	 {CB_CONTENT,">PROFILE>REPGROUP>DIVE>TIME>HOUR","02",&divestruct.time[0],sprintpad },
	 {CB_CONTENT,">PROFILE>REPGROUP>DIVE>TIME>MINUTE","02",&divestruct.time[3],sprintpad },
	 {CB_CLOSE  ,">PROFILE>REPGROUP>DIVE>TIME","Timestart",&divestruct.time[0],texpriv },
	 {CB_CONTENT,">PROFILE>REPGROUP>DIVE>SURFACEINTERVAL","SurfInt",&myconverts[CV_OFP],texmeconv },
	 {CB_CONTENT,">PROFILE>REPGROUP>DIVE>GASES>MIX>MIXNAME","Gasname",NULL,texme },
	 {CB_CONTENT,">PROFILE>REPGROUP>DIVE>GASES>MIX>TANK>TANKVOLUME","Gasvol",&myconverts[CV_VOL],texmeconv },
	 {CB_CONTENT,">PROFILE>REPGROUP>DIVE>GASES>MIX>TANK>PSTART","Gaspressstart",&myconverts[CV_PRESS],texmeconv },
	 {CB_CONTENT,">PROFILE>REPGROUP>DIVE>GASES>MIX>TANK>PEND","Gaspressend",&myconverts[CV_PRESS],texmeconv },
	 {CB_CONTENT,">PROFILE>REPGROUP>DIVE>GASES>MIX>O2","Gasperco2",NULL,texme },
	 {CB_CONTENT,">PROFILE>REPGROUP>DIVE>GASES>MIX>N2","Gaspercn2",NULL,texme },
	 {CB_CONTENT,">PROFILE>REPGROUP>DIVE>GASES>MIX>HE","Gasperche",NULL,texme },
	 {CB_CONTENT,">PROFILE>REPGROUP>DIVE>SAMPLES>DELTA",NULL,&divestruct.delta,getdouble },
	 {CB_CONTENT,">PROFILE>REPGROUP>DIVE>SAMPLES>D",NULL,&myconverts[CV_LENGTH],getdepth },
	 {CB_CONTENT,">PROFILE>REPGROUP>DIVE>SAMPLES>ALARM",NULL,NULL,getalarm },
	 {CB_CLOSE  ,">PROFILE>REPGROUP>DIVE>SAMPLES",NULL,NULL,showdepth },
	 {CB_CLOSE  ,">PROFILE>REPGROUP>DIVE",NULL,NULL,enddive },
	 {CB_CONTENT,NULL,NULL,NULL }
};

int callback_tex(unsigned char *tagpath,unsigned char *content,int what) {

	int x;
#ifdef DEBUG
	switch(what) {
		case  CB_CONTENT:
		    	printf("%s contains %s.\n",tagpath,content);
		    	break;
		case  CB_OPEN:
		    	printf("%s opens.\n",tagpath);
		    	break;
		case  CB_CLOSE:
		    	printf("%s closes.\n",tagpath);
		    	break;
		case  CB_OPTION:
		    	printf("%s has option %s.\n",tagpath,content);
		    	break;
	}
#endif
	for(x=0;reactions[x].path;x++) {
		if (reactions[x].mode==what&&
		    strcasecmp(reactions[x].path,tagpath)==0) {
			reactions[x].runme(&reactions[x],content);
			break;
		}
	}
	return 0;
}

int trivparse(FILE *parseme, int (*cb)(unsigned char *tp,unsigned char *cont,int what) ) {

	int istag=0;
	int lastclose=0;
	int bpos=0;
	
	unsigned char buffer[4096];
	unsigned char content[4096];

	unsigned char tagpath[4096];
	
	strcpy(tagpath,"");
	strcpy(buffer,"");
	strcpy(content,"");
	
	while(!feof(parseme)) {
		int currchr;
		
		currchr=fgetc(parseme);
		
		if (currchr=='<') {
			if (istag)
				return -1;
			istag=1;
			strcpy(content,buffer);
			bpos=0;buffer[0]=0;
			continue;
		} else if (currchr=='>') {
			if (!istag)
				return -1;
			istag=0;
			if (*buffer=='!') {
//				printf("Have comment %s.\n",buffer);
			} else if (*buffer=='/') {
				/* strip tag from path */
				if (strlen(tagpath)>=strlen(buffer) &&
				    0==strcasecmp(tagpath+strlen(tagpath)-strlen(buffer)+1,buffer+1) &&
				    tagpath[strlen(tagpath)-strlen(buffer)]=='>'
				    ) {
				    	unsigned char *hlp;
				    	if (lastclose) {
				    		for(hlp=content;*hlp;hlp++) {
				    			if (!isspace(*hlp)) {
				    				fprintf(stderr,"Unexpected content between closing tags ignored.\n(%s)\n",content);
				    				break;
				    			}
				    		}
				    	} else {
				    		cb(tagpath,content,CB_CONTENT);
					}
			    		cb(tagpath,NULL,CB_CLOSE);
					tagpath[strlen(tagpath)-strlen(buffer)]='\0';
					lastclose=1;
				} else {
					fprintf(stderr,"File corrupt: no match for %s at end of %s.\n",buffer+1,tagpath);
					return -2;
				}
			} else if (strlen(buffer)&&buffer[strlen(buffer)-1]=='/') {
		    		cb(tagpath,buffer,CB_OPTION);
			} else {
				unsigned char *hlp;
				hlp=buffer;
				while(*hlp) {
					if (isspace(*hlp)) {
						*hlp='\0';
						break;
					}
					hlp++;
				}
				if (strlen(tagpath)+strlen(buffer)+1<sizeof(tagpath)-1) {
					strcat(tagpath,">");
					strcat(tagpath,buffer);
			    		cb(tagpath,NULL,CB_OPEN);
					lastclose=0;
				} 			
			}
			bpos=0;buffer[0]=0;
			continue;
		} else {
			if (bpos<sizeof(buffer)-1) {
				buffer[bpos++]=currchr;
				buffer[bpos]='\0';
			}
		}
	}
	if (*tagpath!='\0') {
		fprintf(stderr,"file ends with open tags: %s\n",tagpath);
		return -3;
	}
	return 0;
}

int main(int argc,char *argv[]) {

	int gorc;

#ifdef DEFAULT_TO_IMPERIAL
	go_imperial();
#else
	go_metric();
#endif

	while(-1!=(gorc=getopt(argc,argv,"im"))) {
		switch(gorc) {
			case 'i':	/* Switch to imperial units */
				go_imperial();
				break;
			case 'm':
				go_metric();
				break;
			default:
				fprintf(stderr,"Usage: %s [-i|-m] <input.udcf >output.tex\n",argv[0]);
				return 1;
		}
	}

	printf( "%% divelog parameter file\n"
		"%%\n"
		"%% generated by %s, (c) 2002 Andreas Beck <becka@bedatec.de>.\n"
		"%%\n"
		"\\documentclass[%s]{divelog}\n\n"
		"\\punchmark\n\n"
		"\\begin{document}\n\n"
		"\\Divenr              {4711}\n"
		"\\Area                {NRW}\n"
		"\\Spot                {Hitdorfer See}\n"
		"\\Buddy               {Andreas Beck}\n\n"
		
		,argv[0],mytexts->language);
	trivparse(stdin,callback_tex);
	initdepth(NULL,NULL);	/* free memory */

	printf(	"\\end{document}\n\n"
		"\\endinput\n"
		"%%\n"
		"%% generated by %s, (c) 2002 Andreas Beck <becka@bedatec.de>.\n"
		"%%\n"
		,argv[0]);
	return 0;
}
